package geniusweb.javavenv;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;

import org.apache.commons.io.FileUtils;

import tudelft.utilities.logging.Reporter;

/**
 * Creates and owns a python venv and handles calls to it. Create using the
 * {@link PythonPartyFactory}. This is preferred use, over PythonCall, because
 * it runs your commands inside a venv. To use this the PYTHON3EXE environment
 * variable must point to a working python3.exe version 3.8.5. The exe CAN be a
 * symlink to the actual exe (at least, on mac it works fine)
 */
public class PythonVenv {
	private final static String TMP = System.getProperty("java.io.tmpdir");
	private final static String GENIUSWEBVENV = "GeniusWebPythonVenvs";
	private final static String PYTHON3EXE = "PYTHON3EXE";
	private final static String PYTHON3 = System.getenv(PYTHON3EXE);
	private final Path venvpath;
	private final Reporter logger;

	private final static boolean IS_WINDOWS = System.getProperty("os.name")
			.toLowerCase().contains("win");
	private final static String VENV_PYTHON_EXE = IS_WINDOWS ? "Scripts\\python"
			: "bin/python3";

	/**
	 * 
	 * @param dirname a preferred dirname. Must be [a-zA-Z_0-9]+. You can remove
	 *                non-word chars eg using dirname.replaceAll("\\W", "");
	 * @param logger  the logger to report errors
	 * @throws InterruptedException if the call does not terminate in time.
	 * @throws IOException          if we can not connect properly to the python
	 *                              process
	 * @throws PythonError          if the python program printed to stderr, eg
	 *                              when an error occured. We try to capture all
	 *                              the error lines as text in the pythonerror
	 *                              object.
	 */
	public PythonVenv(String dirname, Reporter logger)
			throws IOException, InterruptedException, PythonError {
		if (logger == null || dirname == null)
			throw new IllegalArgumentException("args must be not null");
		if (!(dirname.matches("\\w+")))
			throw new IllegalArgumentException(
					"dirname must be [a-zA-Z_0-9]+  but got " + dirname);
		this.venvpath = Paths.get(TMP, GENIUSWEBVENV, dirname);
		this.logger = logger;
		createVenv();
	}

	/**
	 * @see #call(List, int)
	 * @param arg       this arg is just split according to whitespaces in the
	 *                  string. Use {@link #call(List, int)} if your command
	 *                  contains whitespaces.
	 * @param timeoutms the timeout in ms
	 * @return result of the call
	 * @throws InterruptedException if the call does not terminate in time.
	 * @throws IOException          if we can not connect properly to the python
	 *                              process
	 * @throws PythonError          if the python program printed to stderr, eg
	 *                              when an error occured. We try to capture all
	 *                              the error lines as text in the pythonerror
	 *                              object.
	 * @throws InterruptedException if call is interrupted
	 */
	public String call(String arg, int timeoutms)
			throws InterruptedException, IOException, PythonError {
		return call(Arrays.asList(arg.split(" ")), timeoutms);
	}

	/**
	 * @param args      the arguments for the call, as simple string, We do
	 *                  split on " " to get the arguments
	 * @param timeoutms the timeout for the call in milliseconds.
	 * @return the result of the call. The call will use the python3 and the
	 *         settings in the venv. So you can safely dd eg "python -m pip
	 *         install "
	 * @throws InterruptedException if the call does not terminate in time.
	 * @throws IOException          if we can not connect properly to the python
	 *                              process
	 * @throws PythonError          if the python program printed to stderr, eg
	 *                              when an error occured. We try to capture all
	 *                              the error lines as text in the pythonerror
	 *                              object.
	 * @throws InterruptedException if call is interrupted
	 */
	public String call(List<String> args, int timeoutms)
			throws InterruptedException, IOException, PythonError {
		return new Call(
				venvpath.resolve(VENV_PYTHON_EXE).toAbsolutePath().toString(),
				args, timeoutms).run();
	}

	/**
	 * @param args      the call arguments (excluding the python startup
	 *                  command)
	 * @param timeoutms timeout for the call, actually is ignored currently
	 * @return a {@link Pipe}, with given args and timeoutms
	 */
	public Pipe pipe(List<String> args, int timeoutms) {
		return new Pipe(
				venvpath.resolve(VENV_PYTHON_EXE).toAbsolutePath().toString(),
				args, timeoutms, logger);
	}

	/**
	 * This should be called when the pythonVenv is not used anymore. It will
	 * not be deleted automatically because you may want to keep the venv for a
	 * next run without needing to recreate it.
	 */
	public void remove() {
		try {
			FileUtils.deleteDirectory(venvpath.toFile());
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void createVenv()
			throws IOException, InterruptedException, PythonError {
		File venvdir = venvpath.toFile();
		logger.log(Level.INFO, "creating venv at " + venvdir);
		if (PYTHON3 == null)
			throw new IllegalStateException("The environment variable "
					+ PYTHON3EXE + " must be set to point to "
					+ "the python3.8 or higher executable to be used");

		String version = new Call(PYTHON3,
				Arrays.asList("-c", "import sys; print(sys.version)"), 500)
						.run();
		if (!(version.startsWith("3.8") || version.startsWith("3.9"))) {
			throw new IllegalStateException("Python version must be 3.8 or 3.9 "
					+ " but " + PYTHON3EXE + " points to " + version);
		}

		if (venvdir.exists()) {
			// #15 don't reuse, it may be corrupt or outdated.
			FileUtils.deleteDirectory(venvdir);
		}

		if (!venvdir.mkdirs())
			throw new IllegalStateException(
					"Failed to create venv dir at " + venvdir);

		new Call(PYTHON3,
				Arrays.asList("-m", "venv", venvdir.getAbsolutePath()), 20000)
						.run();

	}

}
